import subprocess
import tempfile
from contextlib import contextmanager
from pathlib import Path
from traceback import print_exc

import pytest

from synchronicity.type_stubs import write_stub

helpers_dir = Path(__file__).parent / "type_stub_helpers"
assertion_file = helpers_dir / "e2e_example_type_assertions.py"


class FailedMyPyCheck(Exception):
    def __init__(self, output):
        self.output = output


def run_mypy(input_file):
    p = subprocess.Popen(
        ["mypy", input_file], stderr=subprocess.STDOUT, stdout=subprocess.PIPE
    )
    result_code = p.wait()
    if result_code != 0:
        raise FailedMyPyCheck(p.stdout.read())


@contextmanager
def temp_assertion_file(new_assertion):
    template = assertion_file.read_text()
    setup_code, default_assertions = template.split("# assert start")
    assertion_code = setup_code + new_assertion
    with tempfile.NamedTemporaryFile(
        dir=assertion_file.parent, suffix=".py"
    ) as new_file:
        new_file.write(assertion_code.encode("utf8"))
        new_file.flush()
        try:
            yield new_file.name
        except:
            print(f"Exception when running type assertions on:\n{assertion_code}")
            print_exc()
            raise


@pytest.fixture(scope="session")
def interface_file():
    write_stub("test.type_stub_helpers.e2e_example_export")
    yield


def test_mypy_assertions(interface_file):
    run_mypy(assertion_file)


@pytest.mark.parametrize(
    "failing_assertion,error_matches",
    [
        (
            "e2e_example_export.BlockingFoo(1)",
            'incompatible type "int"; expected "str"',
        ),
        (
            "blocking_foo.some_static()",
            'Missing positional argument "arg" in call to "some_static"',
        ),  # missing argument
        (
            "blocking_foo.some_static(True)",
            'Argument 1 to "some_static" of "BlockingFoo" has incompatible type "bool"',
        ),  # bool instead of str
        (
            "e2e_example_export.listify(123)",
            'Value of type variable "_T_Blocking" of "listify" cannot be "int"',
        ),  #  int does not satisfy the type bound of the typevar (!)
    ],
)
def test_failing_assertion(interface_file, failing_assertion, error_matches):
    # since there appears to be no good way of asserting failing type checks (and skipping to the next assertion)
    # we use the assertion file as a template to insert statements that should fail type checking
    with temp_assertion_file(
        failing_assertion
    ) as custom_file:  # we pass int instead of str
        with pytest.raises(FailedMyPyCheck, match=error_matches):
            run_mypy(custom_file)



from synchronicity import Synchronizer


def test_function_sync():
    s = Synchronizer()

    async def f(x):
        return x**2

    f_blocking = s.create_blocking(f)
    assert f_blocking(42) == 1764



"""
Improvement Ideas:
* Extract this into its own package, not linked to synchronicity, but with good extension plugs?
* Don't use the wrapped synchronicity types directly, and instead emit stubs based on the root implementation types directly (but translated to blocking).
* Let synchronicity emit actual function bodies, to avoid runtime wrapping altogether
"""
import collections
import collections.abc
import contextlib
import importlib
import inspect
import sys
import typing
from pathlib import Path
from typing import TypeVar, Generic
from unittest import mock

import sigtools.specifiers  # type: ignore
from sigtools._signatures import EmptyAnnotation, UpgradedAnnotation  # type: ignore

import synchronicity
from synchronicity import Interface, overload_tracking
from synchronicity.synchronizer import TARGET_INTERFACE_ATTR, SYNCHRONIZER_ATTR


class ReprObj:
    # Hacky repr passthrough object so we can pass verbatim type annotations as partial arguments
    # to generic and have them render correctly through `repr()`, used by inspect.Signature etc.
    def __init__(self, repr: str):
        assert isinstance(repr, str), f"{repr} is not a string!"
        self._repr = repr

    def __repr__(self):
        return self._repr

    def __str__(self):
        return self._repr

    def __call__(self):
        # being a callable gets around some generic's automatic type checking of provided types
        # otherwise we get errors like `provided argument is not a type`
        pass


class StubEmitter:
    def __init__(self, target_module):
        self.target_module = target_module
        self.imports = set()
        self.parts = []
        self._indentation = "    "
        self.global_types = set()
        self.referenced_global_types = set()

    @classmethod
    def from_module(cls, module):
        emitter = cls(module.__name__)
        explicit_members = module.__dict__.get("__all__", [])
        for entity_name, entity in module.__dict__.copy().items():
            if (
                hasattr(entity, "__module__")
                and entity.__module__ != module.__name__
                and entity_name not in explicit_members
            ):
                continue  # skip imported stuff, unless it's explicitly in __all__
            if inspect.isclass(entity):
                emitter.add_class(entity, entity_name)
            elif inspect.isfunction(entity):
                emitter.add_function(entity, entity_name, 0)
            elif isinstance(entity, typing.TypeVar):
                emitter.add_type_var(entity, entity_name)
            elif (
                hasattr(entity, "__class__")
                and getattr(entity.__class__, "__module__", None) == module.__name__
            ):
                # instances of stuff
                emitter.add_variable(entity.__class__, entity_name)

        for varname, annotation in getattr(module, "__annotations__", {}).items():
            emitter.add_variable(annotation, varname)

        return emitter

    def add_variable(self, annotation, name):
        # TODO: evaluate string annotations
        self.parts.append(self._get_var_annotation(name, annotation))

    def add_function(self, func, name, indentation_level=0):
        # adds function source code to module
        self.parts.append(
            self._get_function_source_with_overloads(func, name, indentation_level)
        )

    def _get_translated_class_bases(self, cls):
        # get __orig_bases__ (__bases__ with potential generic args) for any class
        # note that this has to unwrap the class first in case of synchronicity wrappers,
        # since synchronicity classes don't preserve/translate __orig_bases__.
        # (This is due to __init_subclass__ triggering in odd ways for wrapper classes)

        if TARGET_INTERFACE_ATTR in cls.__dict__:
            # get base classes from origin class instead, to preserve potential Generic base classes which are otherwise stripped by synchronicitys wrappers
            synchronizer = cls.__dict__[SYNCHRONIZER_ATTR]
            impl_cls = cls.__dict__[synchronizer._original_attr]
            target_interface = cls.__dict__[TARGET_INTERFACE_ATTR]
            impl_bases = self._get_translated_class_bases(impl_cls)

            retranslated_bases = []
            for impl_base in impl_bases:
                retranslated_bases.append(
                    self._translate_annotation(
                        impl_base, synchronizer, target_interface, cls.__module__
                    )
                )

            return tuple(retranslated_bases)

        # the case that the annotation is a Generic base class, but *not* a synchronicity wrapped one
        bases = []
        for b in cls.__dict__.get("__orig_bases__", cls.__bases__):
            bases.append(self._translate_global_annotation(b, cls))
        return bases

    def add_class(self, cls, name):
        self.global_types.add(name)
        bases = []
        for b in self._get_translated_class_bases(cls):
            if b is not object:
                bases.append(self._formatannotation(b))

        bases_str = "" if not bases else "(" + ", ".join(bases) + ")"
        decl = f"class {name}{bases_str}:"
        var_annotations = []
        methods = []

        annotations = cls.__dict__.get("__annotations__", {})
        annotations = {
            k: self._translate_global_annotation(annotation, cls)
            for k, annotation in annotations.items()
        }

        body_indent_level = 1
        body_indent = self._indent(body_indent_level)

        for varname, annotation in annotations.items():
            var_annotations.append(
                f"{body_indent}{self._get_var_annotation(varname, annotation)}"
            )
        if var_annotations:
            var_annotations.append(
                ""
            )  # formatting ocd - add an extra newline after var annotations

        for entity_name, entity in cls.__dict__.items():
            if inspect.isfunction(entity):
                methods.append(
                    self._get_function_source_with_overloads(
                        entity, entity_name, body_indent_level
                    )
                )

            elif isinstance(entity, classmethod):
                methods.append(
                    f"{body_indent}@classmethod\n{self._get_function_source_with_overloads(entity.__func__, entity_name, body_indent_level)}"
                )

            elif isinstance(entity, staticmethod):
                methods.append(
                    f"{body_indent}@staticmethod\n{self._get_function_source_with_overloads(entity.__func__, entity_name, body_indent_level)}"
                )

            elif isinstance(entity, property):
                methods.append(
                    f"{body_indent}@property\n{self._get_function_source_with_overloads(entity.fget, entity_name, body_indent_level)}"
                )

        padding = [] if var_annotations or methods else [f"{body_indent}..."]
        self.parts.append(
            "\n".join(
                [
                    decl,
                    *var_annotations,
                    *methods,
                    *padding,
                ]
            )
        )

    def add_type_var(self, type_var, name):
        self.imports.add("typing")
        args = [f'"{name}"']
        if type_var.__bound__:
            translated_bound = self._translate_global_annotation(
                type_var.__bound__, type_var
            )
            str_annotation = self._formatannotation(translated_bound)
            args.append(f'bound="{str_annotation}"')
        self.global_types.add(name)
        self.parts.append(f'{name} = typing.TypeVar({", ".join(args)})')

    def get_source(self):
        missing_types = self.referenced_global_types - self.global_types
        if missing_types:
            print(
                f"WARNING: {self.target_module} missing the following referenced types, expected to be in module"
            )
            for t in missing_types:
                print(t)
        import_src = "\n".join(sorted(f"import {mod}" for mod in self.imports))
        stubs = "\n\n".join(self.parts)
        return f"{import_src}\n\n{stubs}".lstrip()

    def _ensure_import(self, typ):
        # add import for a single type, non-recursive (See _register_imports)
        # also marks the type name as directly referenced if it's part of the target module
        # so we can sanity check
        module = typ.__module__
        if module not in (self.target_module, "builtins"):
            self.imports.add(module)

        if module == self.target_module:
            if not hasattr(typ, "__name__"):
                # weird special case with Generic subclasses in the target module...
                generic_origin = typ.__origin__
                assert issubclass(generic_origin, Generic)  # noqa
                name = generic_origin.__name__
            else:
                name = typ.__name__
            self.referenced_global_types.add(name)

    def _register_imports(self, type_annotation):
        # recursively makes sure a type and any of its type arguments (for generics) are imported
        origin = getattr(type_annotation, "__origin__", None)
        if origin is None:
            # "scalar" base type
            if hasattr(type_annotation, "__module__"):
                self._ensure_import(type_annotation)
            return

        self._ensure_import(type_annotation)  # import the generic itself's module
        for arg in getattr(type_annotation, "__args__", ()):
            self._register_imports(arg)

    def _translate_global_annotation(self, annotation, source_class_or_function):
        # convenience wrapper for _translate_annotation when the translated entity itself
        # determines eval scope and synchronizer target

        # infers synchronizer, target and home_module from an entity (class, function) containing the annotation
        synchronicity_target_interface = getattr(
            source_class_or_function, TARGET_INTERFACE_ATTR, None
        )
        synchronizer = getattr(source_class_or_function, SYNCHRONIZER_ATTR, None)
        if synchronizer:
            home_module = getattr(
                source_class_or_function, synchronizer._original_attr
            ).__module__
        else:
            home_module = source_class_or_function.__module__

        return self._translate_annotation(
            annotation, synchronizer, synchronicity_target_interface, home_module
        )

    def _translate_annotation(
        self,
        annotation,
        synchronizer: typing.Optional[synchronicity.Synchronizer],
        synchronicity_target_interface: typing.Optional[Interface],
        home_module: typing.Optional[str],
    ):
        """
        Takes an annotation (type, generic, typevar, forward ref) and applies recursively (in case of generics):
        * eval for string annotations (importing `home_module` to be used as namespace)
        * re-mapping of the annotation to the correct synchronicity target (using synchronizer and synchronicity_target_interface)
        * registers imports for all referenced modules
        """
        if isinstance(
            annotation, typing.ForwardRef
        ):  # TypeVars wrap their arguments as ForwardRefs (sometimes?)
            annotation = annotation.__forward_arg__

        if isinstance(annotation, str):
            assert home_module is not None
            mod = importlib.import_module(home_module)
            try:
                annotation = eval(annotation, mod.__dict__)
            except NameError:
                # attempt to import
                guessed_module, name = annotation.rsplit(".", 1)
                exec(f"import {guessed_module}", mod.__dict__)  # import first
                annotation = eval(annotation, mod.__dict__)

        annotation = self._translate_annotation_map_types(
            annotation,
            synchronizer=synchronizer,
            interface=synchronicity_target_interface,
            home_module=home_module,
        )

        self._register_imports(annotation)
        return annotation

    def _translate_annotation_map_types(
        self,
        type_annotation,
        synchronizer: typing.Optional[synchronicity.Synchronizer],
        interface: typing.Optional[Interface],
        home_module: typing.Optional[str] = None,
    ):
        # recursively map a nested type annotation to match the output interface
        origin = getattr(type_annotation, "__origin__", None)
        if origin is None:
            # scalar - if type is synchronicity origin type, use the blocking/async version instead
            if synchronizer:
                return synchronizer._translate_out(type_annotation, interface)
            return type_annotation

        args = getattr(type_annotation, "__args__", [])
        mapped_args = tuple(
            self._translate_annotation(arg, synchronizer, interface, home_module)
            for arg in args
        )
        if interface == Interface.BLOCKING:
            # blocking interface special generic translations:
            if origin == collections.abc.AsyncGenerator:
                return typing.Generator[mapped_args + (None,)]  # type: ignore

            if origin == contextlib.AbstractAsyncContextManager:
                return typing.ContextManager[mapped_args]  # type: ignore

            if origin == collections.abc.AsyncIterable:
                return typing.Iterable[mapped_args]  # type: ignore

            if origin == collections.abc.AsyncIterator:
                return typing.Iterator[mapped_args]  # type: ignore

            if origin == collections.abc.Awaitable:
                return mapped_args[0]

            if origin == collections.abc.Coroutine:
                return mapped_args[2]

        if origin.__module__ not in (
            "typing",
            "collections.abc",
            "contextlib",
        ):  # don't translate built in generics in type annotations, even if they have been synchronicity wrapped
            # for other hierarchy reasons...
            translated_origin = self._translate_annotation(
                origin, synchronizer, interface, home_module
            )
            if translated_origin is not origin:
                # special case for synchronicity-translated generics, due to synchronicitys wrappers not being valid generics
                # kind of ugly as it returns a string representation rather than a type...
                str_args = ", ".join(self._formatannotation(arg) for arg in mapped_args)
                return ReprObj(
                    f"{self._formatannotation(translated_origin)}[{str_args}]"
                )

        return type_annotation.copy_with(mapped_args)

    def _custom_signature(self, func) -> str:
        """
        We use this instead o str(inspect.Signature()) due to a few issues:
        * Generics with None args are incorrectly encoded as NoneType in str(signature)
        * Some names for stdlib module object types omit the module qualification (notably typing)
        * We might have to stringify annotations to support forward/self references
        * General flexibility like not being able to maintain *comments* in the arg declarations if we want to
        * We intentionally do not use follow_wrapped, since it will override runtime-transformed annotations on a wrapper
        * TypeVars default repr is `~T` instead of `origin_module.T` etc.
        """
        sig = sigtools.specifiers.signature(func)

        if sig.upgraded_return_annotation is not EmptyAnnotation:
            return_annotation = sig.upgraded_return_annotation.source_value()
            return_annotation = self._translate_global_annotation(
                return_annotation, func
            )
            sig = sig.replace(
                return_annotation=return_annotation,
                upgraded_return_annotation=UpgradedAnnotation.upgrade(
                    return_annotation, func, None
                ),  # not sure if needed
            )

        new_parameters = []
        for param in sig.parameters.values():
            if param.upgraded_annotation is not EmptyAnnotation:
                raw_annotation = param.upgraded_annotation.source_value()
                raw_annotation = self._translate_global_annotation(raw_annotation, func)
                new_parameters.append(
                    param.replace(
                        annotation=raw_annotation,
                        upgraded_annotation=UpgradedAnnotation.upgrade(
                            raw_annotation, func, param.name
                        ),  # not sure if needed...
                    )
                )
            else:
                new_parameters.append(param)

        sig = sig.replace(parameters=new_parameters)

        # kind of ugly, but this ensures valid formatting of Generics etc, see docstring above
        with mock.patch("inspect.formatannotation", self._formatannotation):
            return str(sig)

    def _get_var_annotation(self, name, annotation):
        # TODO: how to translate annotation here - we don't know the
        self._register_imports(annotation)
        return f"{name}: {self._formatannotation(annotation, None)}"

    def _formatannotation(self, annotation, base_module=None) -> str:
        """modified version of `inspect.formatannotations`
        * Uses verbatim `None` instead of `NoneType` for None-arguments in generic types
        * Doesn't omit `typing.`-module from qualified imports in type names
        * recurses through generic types using ReprObj wrapper
        * ignores base_module (uses self.target_module instead)
        """

        assert (
            base_module is None
        )  # inspect.Signature isn't generally using the base_module arg afaik

        origin = getattr(annotation, "__origin__", None)
        assert not isinstance(
            annotation, typing.ForwardRef
        )  # Forward refs should already have been evaluated!

        if origin is None:
            if annotation == Ellipsis:
                return "..."
            if isinstance(annotation, type) or isinstance(annotation, TypeVar):
                if annotation == None.__class__:  # check for "NoneType"
                    return "None"
                name = (
                    annotation.__qualname__  # type: ignore
                    if hasattr(annotation, "__qualname__")
                    else annotation.__name__
                )
                if annotation.__module__ in ("builtins", self.target_module):
                    return name
                return annotation.__module__ + "." + name
            return repr(annotation)
        # generic:
        args = getattr(annotation, "__args__", ())

        formatted_annotation = str(
            annotation.copy_with(
                # ellipsis (...) needs to be passed as is, or it will be reformatted
                tuple(
                    ReprObj(self._formatannotation(arg))
                    if arg != Ellipsis
                    else Ellipsis
                    for arg in args
                )
            )
        )
        # this is a bit ugly, but gets rid of incorrect module qualification of Generic subclasses:
        # TODO: find a better way...
        if formatted_annotation.startswith(self.target_module + "."):
            return formatted_annotation.split(self.target_module + ".", 1)[1]
        return formatted_annotation

    def _indent(self, level):
        return level * self._indentation

    def _get_function_source_with_overloads(
        self, func, name, indentation_level=0
    ) -> str:
        signature_indent = self._indent(indentation_level)
        body_indent = self._indent(indentation_level + 1)
        parts = []

        interface = func.__dict__.get(TARGET_INTERFACE_ATTR)
        synchronizer = func.__dict__.get(SYNCHRONIZER_ATTR)
        if interface:
            root_func = func.__dict__[SYNCHRONIZER_ATTR]._translate_in(func)
        else:
            root_func = func

        for overload_func in overload_tracking.get_overloads(root_func):
            self._ensure_import(typing.overload)
            parts.append(f"{signature_indent}@typing.overload")
            if interface:
                overload_func = synchronizer._wrap(overload_func, interface, name=name)

            parts.append(
                self._get_function_source(
                    overload_func, name, signature_indent, body_indent
                )
            )

        parts.append(
            self._get_function_source(func, name, signature_indent, body_indent)
        )
        return "\n".join(parts)

    def _get_function_source(
        self, func, name, signature_indent: str, body_indent: str
    ) -> str:
        async_prefix = ""
        if inspect.iscoroutinefunction(func):
            # note: async prefix should not be used for annotated abstract/stub *async generators*, so we don't check for inspect.isasyncgenfunction
            # since they contain no yield keyword, and would otherwise indicate an awaitable that returns an async generator to static type checkers
            async_prefix = "async "

        signature = self._custom_signature(func)

        return "\n".join(
            [
                f"{signature_indent}{async_prefix}def {name}{signature}:",
                f"{body_indent}...",
                "",
            ]
        )


def write_stub(module_path: str):
    mod = importlib.import_module(module_path)
    assert mod.__file__ is not None
    emitter = StubEmitter.from_module(mod)
    source = emitter.get_source()
    stub_path = Path(mod.__file__).with_suffix(".pyi")
    stub_path.write_text(source)
    return stub_path


if __name__ == "__main__":
    with overload_tracking.patched_overload():
        for module_path in sys.argv[1:]:
            out_path = write_stub(module_path)
            print(f"Wrote {out_path}")



import inspect

from synchronicity import Synchronizer


def f(x):
    return x**2


def test_multiwrap_warning(recwarn):
    s = Synchronizer(multiwrap_warning=True)
    f_s = s.create_blocking(f)
    assert f_s(42) == 1764
    assert len(recwarn) == 0
    f_s_s = s.create_blocking(f_s)
    assert f_s_s(42) == 1764
    assert len(recwarn) == 1


def test_multiwrap_no_warning(recwarn):
    s = Synchronizer()
    f_s = s.create_blocking(f)
    assert f_s(42) == 1764
    f_s_s = s.create_blocking(f_s)
    assert f_s_s(42) == 1764
    assert len(recwarn) == 0


async def asyncgen():
    yield 42


async def returns_asyncgen():
    return asyncgen()


def test_check_double_wrapped(recwarn):
    s = Synchronizer()
    assert len(recwarn) == 0
    ret = s.create_blocking(returns_asyncgen)()
    assert inspect.isasyncgen(ret)
    assert len(recwarn) == 1



"""Utility for monkey patching typing.overload to allow run time retrieval overloads

Requires any @typing.overload to happen within the patched_overload contextmanager, e.g.:

```python
with patched_overload():
    # the following could be imported from some other module (as long as it wasn't already loaded), or inlined:

    @typing.overload
    def foo(a: int) -> float:
        ...

    def foo(a: typing.Union[bool, int]) -> typing.Union[bool, float]:
        if isinstance(a, bool):
            return a
        return float(a)

foo_overloads = get_overloads(foo)  # returns reference to the overloads of foo (the int -> float one in this case) in the order they are declared
"""
import contextlib
import typing
from unittest import mock

overloads: typing.Dict[typing.Tuple[str, str], typing.List] = {}
original_overload = typing.overload


class Untrackable(Exception):
    pass


def _function_locator(f):
    try:
        return (f.__module__, f.__qualname__)
    except AttributeError:
        raise Untrackable()  # TODO(elias): handle descriptors like classmethod


def _tracking_overload(f):
    # hacky thing to track all typing.overload declarations
    global overloads, original_overload
    try:
        locator = _function_locator(f)
        overloads.setdefault(locator, []).append(f)
    except Untrackable:
        print(f"WARNING: can't track overloads for {f}")

    return original_overload(f)


@contextlib.contextmanager
def patched_overload():
    with mock.patch("typing.overload", _tracking_overload):
        yield


def get_overloads(f) -> typing.List:
    try:
        return overloads.get(_function_locator(f), [])
    except Untrackable:
        return []



import asyncio
import concurrent
import inspect
import pytest
import time

from synchronicity import Synchronizer

SLEEP_DELAY = 0.1


class CustomException(Exception):
    pass


async def f_raises():
    await asyncio.sleep(0.1)
    raise CustomException("something failed")


def test_function_raises_sync():
    s = Synchronizer()
    t0 = time.time()
    with pytest.raises(CustomException):
        f_raises_s = s.create_blocking(f_raises)
        f_raises_s()
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


def test_function_raises_sync_futures():
    s = Synchronizer()
    t0 = time.time()
    f_raises_s = s.create_blocking(f_raises)
    fut = f_raises_s(_future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException):
        fut.result()
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_function_raises_async():
    s = Synchronizer()
    t0 = time.time()
    f_raises_s = s.create_async(f_raises)
    coro = f_raises_s()
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException):
        await coro
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


async def f_raises_baseexc():
    await asyncio.sleep(0.1)
    raise KeyboardInterrupt


def test_function_raises_baseexc_sync():
    s = Synchronizer()
    t0 = time.time()
    with pytest.raises(BaseException):
        f_raises_baseexc_s = s.create_blocking(f_raises_baseexc)
        f_raises_baseexc_s()
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


def f_raises_syncwrap():
    return f_raises()  # returns a coro


@pytest.mark.asyncio
async def test_function_raises_async_syncwrap():
    s = Synchronizer()
    t0 = time.time()
    f_raises_syncwrap_s = s.create_async(f_raises_syncwrap)
    coro = f_raises_syncwrap_s()
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException):
        await coro
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY



# this code is only meant to be "running" through mypy and not an actual python interpreter!
import typing

from test.type_stub_helpers import e2e_example_export
from typing_extensions import assert_type

blocking_foo = e2e_example_export.BlockingFoo("hello")

# assert start
assert_type(blocking_foo, e2e_example_export.BlockingFoo)

assert_type(blocking_foo.getarg(), str)
assert_type(blocking_foo.gen(), typing.Generator[int, None, None])

assert_type(
    e2e_example_export.some_instance, typing.Optional[e2e_example_export.BlockingFoo]
)

assert_type(blocking_foo.some_static("foo"), float)

assert_type(
    e2e_example_export.BlockingFoo.clone(blocking_foo), e2e_example_export.BlockingFoo
)

assert_type(blocking_foo.singleton, e2e_example_export.BlockingFoo)


assert_type(
    e2e_example_export.listify(blocking_foo),
    typing.List[e2e_example_export.BlockingFoo],
)



import asyncio
import pytest
from typing import Dict, Any

from synchronicity import Synchronizer


def test_getattr():
    s = Synchronizer()

    class Foo:
        _attrs: Dict[str, Any]

        def __init__(self):
            self._attrs = {}

        async def __getattr__(self, k):
            await asyncio.sleep(0.01)
            return self._attrs[k]

        def __setattr__(self, k, v):
            if k in self.__annotations__:
                # Only needed because the constructor sets _attrs
                self.__dict__[k] = v
            else:
                self._attrs[k] = v

        @property
        def z(self):
            return self._attrs["x"]

        @staticmethod
        def make_foo():
            return Foo()

    foo = Foo()
    foo.x = 42
    assert asyncio.run(foo.x) == 42
    with pytest.raises(KeyError):
        asyncio.run(foo.y)
    assert foo.z == 42

    BlockingFoo = s.create_blocking(Foo)

    blocking_foo = BlockingFoo()
    blocking_foo.x = 42
    assert blocking_foo.x == 42
    with pytest.raises(KeyError):
        blocking_foo.y
    assert blocking_foo.z == 42

    blocking_foo = BlockingFoo.make_foo()
    assert isinstance(blocking_foo, BlockingFoo)

    AsyncFoo = s.create_async(Foo)
    async_foo = AsyncFoo()
    async_foo.x = 42
    assert asyncio.run(async_foo.x) == 42
    assert async_foo.z == 42



import inspect
import typing

import synchronicity
from synchronicity import Interface, async_wrap
from synchronicity.async_wrap import wraps_by_interface


def test_wrap_corofunc_using_async():
    async def foo():
        pass

    @wraps_by_interface(Interface.ASYNC, foo)
    async def bar():
        pass

    assert inspect.iscoroutinefunction(bar)


def test_wrap_corofunc_using_non_async():
    async def foo():
        pass

    @wraps_by_interface(Interface.ASYNC, foo)
    def bar():
        pass

    assert inspect.iscoroutinefunction(bar)


def test_wrap_asynccontextmanager_annotations():
    @async_wrap.asynccontextmanager  # this would not work with contextlib.asynccontextmanager
    async def foo() -> typing.AsyncGenerator[int, None]:
        pass

    assert foo.__annotations__["return"] == typing.AsyncContextManager[int]


def test_wrap_staticmethod():
    class Foo:
        @staticmethod
        async def a_static_method() -> typing.Awaitable[str]:
            async def wrapped():
                return "hello"

            return wrapped()

    synchronizer = synchronicity.Synchronizer()
    BlockingFoo = synchronizer.create_blocking(Foo)
    AsyncFoo = synchronizer.create_async(Foo)

    assert isinstance(BlockingFoo.__dict__["a_static_method"], staticmethod)
    assert isinstance(AsyncFoo.__dict__["a_static_method"], staticmethod)

    assert inspect.iscoroutinefunction(AsyncFoo.__dict__["a_static_method"].__func__)
    assert not inspect.iscoroutinefunction(
        BlockingFoo.__dict__["a_static_method"].__func__
    )



import pytest
import traceback

from synchronicity import Synchronizer


class CustomException(Exception):
    pass


async def f():
    raise CustomException("boom!")


async def f_baseexc():
    raise KeyboardInterrupt


async def gen():
    raise CustomException("gen boom!")
    yield


def check_traceback(exc):
    tb = exc.__traceback__
    n_outside = 0
    for frame in traceback.extract_tb(tb):
        if frame.filename != __file__:
            n_outside += 1

    # Let's allow for at most 1 entry outside of this file
    if n_outside > 1:
        for frame in traceback.extract_tb(tb):
            print(frame.filename, frame.lineno, frame.line)
        raise Exception(f"Got {n_outside} frames outside user code, expected 1")


def test_sync_to_async():
    s = Synchronizer()
    f_s = s.create_blocking(f)
    with pytest.raises(CustomException) as excinfo:
        f_s()
    check_traceback(excinfo.value)


@pytest.mark.asyncio
async def test_async_to_async():
    s = Synchronizer()
    f_s = s.create_async(f)
    with pytest.raises(CustomException) as excinfo:
        await f_s()
    check_traceback(excinfo.value)


def test_sync_to_async_gen():
    s = Synchronizer()
    gen_s = s.create_blocking(gen)
    with pytest.raises(CustomException) as excinfo:
        for x in gen_s():
            pass
    check_traceback(excinfo.value)


@pytest.mark.asyncio
async def test_async_to_async_gen():
    s = Synchronizer()
    gen_s = s.create_async(gen)
    with pytest.raises(CustomException) as excinfo:
        async for x in gen_s():
            pass
    check_traceback(excinfo.value)


@pytest.mark.skip(reason="This one will be much easier to fix once AUTODETECT is gone")
def test_sync_to_async_ctx_mgr():
    s = Synchronizer()
    ctx_mgr = s.create_blocking(s.asynccontextmanager(gen))
    with pytest.raises(CustomException) as excinfo:
        with ctx_mgr():
            pass
    check_traceback(excinfo.value)


@pytest.mark.skip(reason="This one will be much easier to fix once AUTODETECT is gone")
@pytest.mark.asyncio
async def test_async_to_async_ctx_mgr():
    s = Synchronizer()
    ctx_mgr = s.create_async(s.asynccontextmanager(gen))
    with pytest.raises(CustomException) as excinfo:
        async with ctx_mgr():
            pass
    check_traceback(excinfo.value)


def test_recursive():
    s = Synchronizer()

    async def f(n):
        if n == 0:
            raise CustomException("boom!")
        else:
            return await f(n - 1)

    with pytest.raises(CustomException) as excinfo:
        f_blocking = s.create_blocking(f)
        f_blocking(10)
    check_traceback(excinfo.value)



import functools
import typing

import synchronicity
from synchronicity import overload_tracking
from synchronicity.type_stubs import StubEmitter


def noop():
    pass


def arg_no_anno(arg1):
    pass


def scalar_args(arg1: str, arg2: int) -> float:
    pass


from .type_stub_helpers import some_mod


def generic_other_module_arg(arg: typing.List[some_mod.Foo]):
    pass


async def async_func() -> str:
    return "hello"


def _function_source(func, target_module=__name__):
    stub_emitter = StubEmitter(target_module)
    stub_emitter.add_function(func, func.__name__)
    return stub_emitter.get_source()


def _class_source(cls, target_module=__name__):
    stub_emitter = StubEmitter(target_module)
    stub_emitter.add_class(cls, cls.__name__)
    return stub_emitter.get_source()


def test_function_basics():
    assert _function_source(noop) == "def noop():\n    ...\n"
    assert _function_source(arg_no_anno) == "def arg_no_anno(arg1):\n    ...\n"
    assert (
        _function_source(scalar_args)
        == "def scalar_args(arg1: str, arg2: int) -> float:\n    ...\n"
    )


def test_function_with_imports():
    assert (
        _function_source(generic_other_module_arg, target_module="dummy")
        == """import test.type_stub_helpers.some_mod
import typing

def generic_other_module_arg(arg: typing.List[test.type_stub_helpers.some_mod.Foo]):
    ...
"""
    )


def test_async_func():
    assert _function_source(async_func) == "async def async_func() -> str:\n    ...\n"


def test_async_gen():
    async def async_gen() -> typing.AsyncGenerator[int, None]:
        yield 0

    assert (
        _function_source(async_gen)
        == "import typing\n\ndef async_gen() -> typing.AsyncGenerator[int, None]:\n    ...\n"
    )

    def weird_async_gen() -> typing.AsyncGenerator[int, None]:
        # non-async function that returns an async generator
        async def gen():
            yield 0

        return gen()

    assert (
        _function_source(weird_async_gen)
        == "import typing\n\ndef weird_async_gen() -> typing.AsyncGenerator[int, None]:\n    ...\n"
    )

    async def it() -> typing.AsyncIterator[str]:  # this is the/a correct annotation
        yield "hello"

    src = _function_source(it)
    assert "yield" not in src
    # since the yield keyword is removed in a type stub, the async prefix needs to be removed as well to avoid "double asyncness" (while keeping the remaining annotation)
    assert "async" not in src
    assert "def it() -> typing.AsyncIterator[str]:" in src


class MixedClass:
    class_var: str

    def some_method(self) -> bool:
        return False

    @classmethod
    def some_class_method(cls) -> int:
        return 1

    @staticmethod
    def some_staticmethod() -> float:
        return 0.0

    @property
    def some_property(self) -> str:
        return ""


def test_class_generation():
    emitter = StubEmitter(__name__)
    emitter.add_class(MixedClass, "MixedClass")
    source = emitter.get_source()
    last_assertion_location = None

    def assert_in_after_last(search_string):
        nonlocal last_assertion_location
        assert search_string in source
        if last_assertion_location is not None:
            new_location = source.find(search_string)
            assert new_location > last_assertion_location
            last_assertion_location = new_location

    indent = "    "
    assert_in_after_last("class MixedClass:")
    assert_in_after_last(f"{indent}class_var: str")
    assert_in_after_last(f"{indent}class_var: str")
    assert_in_after_last(f"{indent}def some_method(self) -> bool:\n{indent * 2}...")
    assert_in_after_last(
        f"{indent}@classmethod\n{indent}def some_class_method(cls) -> int:\n{indent * 2}..."
    )
    assert_in_after_last(
        f"{indent}@staticmethod\n{indent}def some_staticmethod() -> float:"
    )
    assert_in_after_last(f"{indent}@property\n{indent}def some_property(self) -> str:")


def merged_signature(*sigs):
    sig = sigs[0].copy()
    return sig


def test_wrapped_function_with_new_annotations():
    """A wrapped function (in general, using functools.wraps/partial) would
    have an inspect.signature from the wrapped function by default
    and from the wrapper function is inspect.signature gets the follow_wrapped=True
    option. However, for the best type stub usability, the best would be to combine
    all layers of wrapping, adding any additional arguments or annotations as updates
    to the underlying wrapped function signature.

    This test makes sure we do just that.
    """

    def orig(arg: str):
        pass

    @functools.wraps(orig)
    def wrapper(extra_arg: int, *args, **kwargs):
        orig(*args, **kwargs)

    wrapper.__annotations__.update({"extra_arg": int, "arg": float})
    assert (
        _function_source(wrapper) == "def orig(extra_arg: int, arg: float):\n    ...\n"
    )


class Base:
    def base_method(self) -> str:
        pass


Base.__module__ = "basemod"
Base.__qualname__ = "Base"


class Based(Base):
    def sub(self) -> float:
        pass


def test_base_class_included_and_imported():
    src = _class_source(Based)
    assert "import basemod" in src
    assert "class Based(basemod.Base):" in src
    assert "base_method" not in src  # base class method should not be in emitted stub


def test_typevar():
    T = typing.TypeVar("T")
    T.__module__ = "source_mod"

    def foo(arg: T) -> T:
        pass

    src = _function_source(foo)
    assert "import source_mod" in src
    assert "def foo(arg: source_mod.T) -> source_mod.T" in src


def test_string_annotation():
    stub_emitter = StubEmitter("dummy")
    stub_emitter.add_variable(annotation="Foo", name="some_foo")  # string annotation
    src = stub_emitter.get_source()
    assert 'some_foo: "Foo"' in src or "some_foo: 'Foo'" in src


class Forwarder:
    def foo(self) -> typing.Optional["Forwardee"]:
        pass


class Forwardee:
    pass


def test_forward_ref():
    # add in the same order here:
    stub = StubEmitter(__name__)
    stub.add_class(Forwarder, "Forwarder")
    stub.add_class(Forwardee, "Forwardee")
    src = stub.get_source()
    assert "class Forwarder:" in src
    assert (
        "def foo(self) -> typing.Union[Forwardee, None]:" in src
    )  # should technically be 'Forwardee', but non-strings seem ok in pure type stubs


class SelfRefFoo:
    def foo(self) -> "SelfRefFoo":
        pass


def test_self_ref():
    src = _class_source(SelfRefFoo)
    assert (
        "def foo(self) -> SelfRefFoo" in src
    )  # should technically be 'Foo' but non-strings seem ok in pure type stubs


class _Foo:
    @staticmethod
    async def clone(foo: "_Foo") -> "_Foo":
        pass


synchronizer = synchronicity.Synchronizer()
Foo = synchronizer.create_blocking(_Foo, "Foo", __name__)


def test_synchronicity_type_translation():
    async def _get_foo() -> _Foo:
        pass

    get_foo = synchronizer.create_blocking(_get_foo, "get_foo", __name__)
    src = _function_source(get_foo)
    assert "def get_foo() -> Foo" in src


def test_synchronicity_self_ref():
    src = _class_source(Foo)
    assert "@staticmethod" in src
    assert "    def clone(foo: Foo) -> Foo" in src


class _WithClassMethod:
    @classmethod
    def classy(cls):
        pass


WithClassMethod = synchronizer.create_blocking(
    _WithClassMethod, "WithClassMethod", __name__
)


def test_synchronicity_classmethod():
    src = _class_source(WithClassMethod)
    assert "@classmethod" in src
    assert "    def classy(cls):" in src


T = typing.TypeVar("T")


class MyGeneric(typing.Generic[T]):
    pass


BlockingGeneric = synchronizer.create_blocking(
    typing.Generic, "BlockingGeneric", __name__
)
BlockingMyGeneric = synchronizer.create_blocking(
    MyGeneric,
    "BlockingMyGeneric",
    __name__,
)


def test_custom_generic():
    # TODO: build out this test a bit, as it currently creates an invalid stub (missing base types)
    class Specific(MyGeneric[str]):
        pass

    src = _class_source(Specific)
    assert "class Specific(MyGeneric[str]):" in src


def test_synchronicity_generic_subclass():
    class Specific(MyGeneric[str]):
        pass

    assert Specific.__bases__ == (MyGeneric,)
    assert Specific.__orig_bases__ == (MyGeneric[str],)

    BlockingSpecific = synchronizer.create_blocking(
        Specific, "BlockingSpecific", __name__
    )
    src = _class_source(BlockingSpecific)
    assert "class BlockingSpecific(BlockingMyGeneric[str]):" in src

    async def foo_impl(bar: MyGeneric[str]):
        pass

    foo = synchronizer.create_blocking(foo_impl, "foo")
    src = _function_source(foo)
    assert "def foo(bar: BlockingMyGeneric[str]):" in src


_B = typing.TypeVar("_B", bound="str")

B = synchronizer.create_blocking(
    _B, "B", __name__
)  # only strictly needed if the bound is a synchronicity implementation type


def _ident(b: _B) -> _B:
    return b


ident = synchronizer.create_blocking(_ident, "ident", __name__)


def test_translated_bound_type_vars():
    emitter = StubEmitter(__name__)
    emitter.add_type_var(B, "B")
    emitter.add_function(ident, "ident")
    src = emitter.get_source()
    assert 'B = typing.TypeVar("B", bound="str")' in src
    assert "def ident(b: B) -> B" in src


def test_ellipsis():
    def foo() -> typing.Callable[..., typing.Any]:
        pass

    src = _function_source(foo)
    assert "-> typing.Callable[..., typing.Any]" in src


def test_overloads_unwrapped_functions():
    with overload_tracking.patched_overload():

        @typing.overload
        def foo(a: int) -> float:
            ...

        def foo(a: typing.Union[bool, int]) -> typing.Union[bool, float]:
            if isinstance(a, bool):
                return a
            return float(a)

        blocking_foo = synchronizer.create_blocking(foo, "blocking_foo")

        src = _function_source(foo)
        assert "@typing.overload\ndef foo(a: int) -> float:" in src
        assert (
            "def foo(a: typing.Union[bool, int]) -> typing.Union[bool, float]:" in src
        )

        src = _function_source(blocking_foo)
        assert "@typing.overload\ndef blocking_foo(a: int) -> float:" in src
        assert (
            "def blocking_foo(a: typing.Union[bool, int]) -> typing.Union[bool, float]:"
            in src
        )



import collections.abc
import contextlib
import functools
import inspect
import typing

from .exceptions import UserCodeException
from .interface import Interface
from contextlib import asynccontextmanager as _asynccontextmanager


def wraps_by_interface(interface: Interface, func):
    """Like functools.wraps but maintains `inspect.iscoroutinefunction` and allows custom type annotations overrides

    Use this when the wrapper function is non-async but returns the coroutine resulting
    from calling the underlying wrapped `func`. This will make sure that the wrapper
    is still an async function in that case, and can be inspected as such.

    Note: Does not forward async generator information other than explicit annotations
    """
    if inspect.iscoroutinefunction(func) and interface == Interface.ASYNC:

        def asyncfunc_deco(user_wrapper):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                try:
                    return await user_wrapper(*args, **kwargs)
                except UserCodeException as uc_exc:
                    raise uc_exc.exc from None

            return wrapper

        return asyncfunc_deco
    else:
        return functools.wraps(func)


YIELD_TYPE = typing.TypeVar("YIELD_TYPE")
SEND_TYPE = typing.TypeVar("SEND_TYPE")


def asynccontextmanager(
    f: typing.Callable[..., typing.AsyncGenerator[YIELD_TYPE, SEND_TYPE]]
) -> typing.Callable[[], typing.AsyncContextManager[YIELD_TYPE]]:
    """Wrapper around contextlib.asynccontextmanager that sets correct type annotations

    The standard library one doesn't
    """
    acm_factory: typing.Callable[
        ..., typing.AsyncContextManager[YIELD_TYPE]
    ] = _asynccontextmanager(f)

    old_ret = acm_factory.__annotations__.pop("return", None)
    if old_ret is not None:
        if old_ret.__origin__ in [
            collections.abc.AsyncGenerator,
            collections.abc.AsyncIterator,
            collections.abc.AsyncIterator,
        ]:
            acm_factory.__annotations__["return"] = typing.AsyncContextManager[
                old_ret.__args__[0]  # type: ignore
            ]
        elif old_ret.__origin__ == contextlib.AbstractAsyncContextManager:
            # if the standard lib fixes the annotations in the future, lets not break it...
            return acm_factory
    else:
        raise ValueError(
            "To use the fixed @asynccontextmanager, make sure to properly annotate your wrapped function as an AsyncGenerator"
        )

    return acm_factory



import asyncio
from synchronicity import Synchronizer


async def run():
    try:
        while True:
            print("running")
            await asyncio.sleep(0.3)
    except asyncio.CancelledError:
        print("cancelled")
        raise
    finally:
        print("stopping")
        await asyncio.sleep(0.1)
        print("exiting")


s = Synchronizer()

try:
    s.create_blocking(run)()
except KeyboardInterrupt:
    pass



from typing import AsyncGenerator, TypeVar, List


class _Foo:
    singleton: "_Foo"

    def __init__(self, arg: str):
        self.arg = arg

    async def getarg(self) -> str:
        return self.arg

    async def gen(self) -> AsyncGenerator[int, None]:
        yield 1

    @staticmethod
    def some_static(arg: str) -> float:
        pass

    @classmethod
    def clone(cls, foo: "_Foo") -> "_Foo":  # self ref
        pass


_T = TypeVar("_T", bound=_Foo)


def _listify(t: _T) -> List[_T]:
    return t



from synchronicity import Synchronizer


def test_is_synchronized():
    s = Synchronizer()

    class Foo:
        pass

    BlockingFoo = s.create_blocking(Foo)
    assert s.is_synchronized(Foo) is False
    assert s.is_synchronized(BlockingFoo) is True



import subprocess
import os
import sys


def test_gevent():
    # Run it in a separate process because gevent modifies a lot of modules
    fn = os.path.join(os.path.dirname(__file__), "_gevent.py")
    ret = subprocess.run(
        [sys.executable, fn], stdout=sys.stdout, stderr=sys.stderr, timeout=5
    )
    assert ret.returncode == 0



from synchronicity import Synchronizer


def test_docs():
    s = Synchronizer()

    class Foo:
        def __init__(self):
            """init docs"""
            self._attrs = {}

        def bar(self):
            """bar docs"""

    foo = Foo()
    assert foo.__init__.__doc__ == "init docs"
    assert foo.bar.__doc__ == "bar docs"

    BlockingFoo = s.create_blocking(Foo)
    blocking_foo = BlockingFoo()
    assert blocking_foo.__init__.__doc__ == "init docs"
    assert blocking_foo.bar.__doc__ == "bar docs"

    AsyncFoo = s.create_async(Foo)
    async_foo = AsyncFoo()
    assert async_foo.__init__.__doc__ == "init docs"
    assert async_foo.bar.__doc__ == "bar docs"



import subprocess
import os
import signal
import sys


def test_shutdown():
    # We run it in a separate process so we can simulate interrupting it
    fn = os.path.join(os.path.dirname(__file__), "_shutdown.py")
    p = subprocess.Popen(
        [sys.executable, fn],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env={"PYTHONUNBUFFERED": "1"},
    )
    for i in range(3):  # this number doesn't matter, it's a while loop
        assert p.stdout.readline() == b"running\n"
    p.send_signal(signal.SIGINT)
    assert p.stdout.readline() == b"cancelled\n"
    assert p.stdout.readline() == b"stopping\n"
    assert p.stdout.readline() == b"exiting\n"
    stderr_content = p.stderr.read()
    assert b"Traceback" not in stderr_content



# This file creates wrapped entities for async implementation in e2e_example_impl.py
# This file is then used as input for generating type stubs
from typing import Optional

import synchronicity
from . import e2e_example_impl

synchronizer = synchronicity.Synchronizer()
BlockingFoo = synchronizer.create_blocking(
    e2e_example_impl._Foo, "BlockingFoo", __name__
)

some_instance: Optional[BlockingFoo] = None

_T_Blocking = synchronizer.create_blocking(
    e2e_example_impl._T, "_T_Blocking", __name__
)  # synchronize the TypeVar to support translation of bounds
listify = synchronizer.create_blocking(e2e_example_impl._listify, "listify", __name__)



from contextlib import asynccontextmanager
import pytest

from synchronicity import Synchronizer


async def noop():
    pass


async def error():
    raise Exception("problem")


s = Synchronizer()


class Resource:
    def __init__(self):
        self.state = "none"

    def get_state(self):
        return self.state

    @asynccontextmanager
    async def wrap(self):
        self.state = "entered"
        try:
            yield
        finally:
            self.state = "exited"

    @asynccontextmanager
    async def wrap_yield_twice(self):
        yield
        yield

    @asynccontextmanager
    async def wrap_never_yield(self):
        if False:
            yield


def test_asynccontextmanager_sync():
    r = s.create_blocking(Resource)()
    assert r.get_state() == "none"
    with r.wrap():
        assert r.get_state() == "entered"
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_async():
    r = s.create_async(Resource)()
    assert r.get_state() == "none"
    async with r.wrap():
        assert r.get_state() == "entered"
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_async_raise():
    r = s.create_async(Resource)()
    assert r.get_state() == "none"
    with pytest.raises(Exception):
        async with r.wrap():
            assert r.get_state() == "entered"
            raise Exception("boom")
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_yield_twice():
    r = s.create_async(Resource)()
    with pytest.raises(RuntimeError):
        async with r.wrap_yield_twice():
            pass


@pytest.mark.asyncio
async def test_asynccontextmanager_never_yield():
    r = s.create_async(Resource)()
    with pytest.raises(RuntimeError):
        async with r.wrap_never_yield():
            pass


@pytest.mark.asyncio
async def test_asynccontextmanager_nested():
    s = Synchronizer()
    finally_blocks = []

    @s.create_async
    @asynccontextmanager
    async def a():
        try:
            yield "foo"
        finally:
            finally_blocks.append("A")

    @s.create_async
    @asynccontextmanager
    async def b():
        async with a() as it:
            try:
                yield it
            finally:
                finally_blocks.append("B")

    with pytest.raises(BaseException):
        async with b():
            raise BaseException("boom!")

    assert finally_blocks == ["B", "A"]


@pytest.mark.asyncio
async def test_asynccontextmanager_with_in_async():
    r = s.create_async(Resource)()
    with pytest.raises(AttributeError):
        with r.wrap():
            pass



import asyncio
import concurrent.futures
import time

from synchronicity import Synchronizer


def test_start_loop():
    # Make sure there's no race condition in _start_loop
    s = Synchronizer()

    with concurrent.futures.ThreadPoolExecutor() as executor:
        ret = list(executor.map(lambda i: s._start_loop(), range(1000)))

    assert len(set(ret)) == 1
    assert isinstance(ret[0], asyncio.AbstractEventLoop)


async def f(i):
    await asyncio.sleep(1.0)
    return i**2


def test_multithreaded(n_threads=20):
    s = Synchronizer()

    f_s = s.create_blocking(f)

    t0 = time.time()
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_threads) as executor:
        ret = list(executor.map(f_s, range(n_threads)))
    assert 1.0 <= time.time() - t0 < 1.2
    assert ret == [i**2 for i in range(n_threads)]



import asyncio
import concurrent.futures
import inspect
import pytest
import time

from synchronicity import Synchronizer, Interface

SLEEP_DELAY = 0.1


async def f(x):
    await asyncio.sleep(SLEEP_DELAY)
    return x**2


async def f2(fn, x):
    return await fn(x)


def test_function_sync():
    s = Synchronizer()
    t0 = time.time()
    f_s = s.create_blocking(f)
    assert f_s.__name__ == "blocking_f"
    ret = f_s(42)
    assert ret == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


def test_function_sync_future():
    s = Synchronizer()
    t0 = time.time()
    f_s = s.create_blocking(f)
    assert f_s.__name__ == "blocking_f"
    fut = f_s(42, _future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert fut.result() == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_function_async():
    s = Synchronizer()
    t0 = time.time()
    f_s = s.create_async(f)
    assert f_s.__name__ == "async_f"
    coro = f_s(42)
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    assert await coro == 1764
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    # Make sure the same-loop calls work
    f2_s = s.create_async(f2)
    assert f2_s.__name__ == "async_f2"
    coro = f2_s(f_s, 42)
    assert await coro == 1764

    # Make sure cross-loop calls work
    s2 = Synchronizer()
    f2_s2 = s2.create_async(f2)
    assert f2_s2.__name__ == "async_f2"
    coro = f2_s2(f_s, 42)
    assert await coro == 1764


@pytest.mark.asyncio
async def test_function_async_block_event_loop():
    s = Synchronizer()

    async def spinlock():
        # This blocks the event loop, but not the main event loop
        time.sleep(SLEEP_DELAY)

    spinlock_s = s.create_async(spinlock)
    spinlock_coro = spinlock_s()
    sleep_coro = asyncio.sleep(SLEEP_DELAY)

    t0 = time.time()
    await asyncio.gather(spinlock_coro, sleep_coro)
    assert SLEEP_DELAY <= time.time() - t0 < 2 * SLEEP_DELAY


def test_function_many_parallel_sync():
    s = Synchronizer()
    g = s.create_blocking(f)
    t0 = time.time()
    rets = [g(i) for i in range(10)]  # Will resolve serially
    assert len(rets) * SLEEP_DELAY < time.time() - t0 < (len(rets) + 1) * SLEEP_DELAY


def test_function_many_parallel_sync_futures():
    s = Synchronizer()
    g = s.create_blocking(f)
    t0 = time.time()
    futs = [g(i, _future=True) for i in range(100)]
    assert isinstance(futs[0], concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert [fut.result() for fut in futs] == [z**2 for z in range(100)]
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_function_many_parallel_async():
    s = Synchronizer()
    g = s.create_async(f)
    t0 = time.time()
    coros = [g(i) for i in range(100)]
    assert inspect.iscoroutine(coros[0])
    assert time.time() - t0 < SLEEP_DELAY
    assert await asyncio.gather(*coros) == [z**2 for z in range(100)]
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY


async def gen(n):
    for i in range(n):
        await asyncio.sleep(0.1)
        yield i


async def gen2(generator, n):
    async for ret in generator(n):
        yield ret


def test_generator_sync():
    s = Synchronizer()
    t0 = time.time()
    gen_s = s.create_blocking(gen)
    it = gen_s(3)
    assert inspect.isgenerator(it)
    assert time.time() - t0 < SLEEP_DELAY
    lst = list(it)
    assert lst == [0, 1, 2]
    assert time.time() - t0 > len(lst) * SLEEP_DELAY


@pytest.mark.asyncio
async def test_generator_async():
    s = Synchronizer()
    t0 = time.time()
    gen_s = s.create_async(gen)
    asyncgen = gen_s(3)
    assert inspect.isasyncgen(asyncgen)
    assert time.time() - t0 < SLEEP_DELAY
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]
    assert time.time() - t0 > len(lst) * SLEEP_DELAY

    # Make sure same-loop calls work
    gen2_s = s.create_async(gen2)
    asyncgen = gen2_s(gen_s, 3)
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]

    # Make sure cross-loop calls work
    s2 = Synchronizer()
    gen2_s2 = s2.create_async(gen2)
    asyncgen = gen2_s2(gen_s, 3)
    lst = [z async for z in asyncgen]
    assert lst == [0, 1, 2]


def test_sync_lambda_returning_coroutine_sync():
    s = Synchronizer()
    t0 = time.time()
    g = s.create_blocking(lambda z: f(z + 1))
    ret = g(42)
    assert ret == 1849
    assert time.time() - t0 > SLEEP_DELAY


def test_sync_lambda_returning_coroutine_sync_futures():
    s = Synchronizer()
    t0 = time.time()
    g = s.create_blocking(lambda z: f(z + 1))
    fut = g(42, _future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.time() - t0 < SLEEP_DELAY
    assert fut.result() == 1849
    assert time.time() - t0 > SLEEP_DELAY


@pytest.mark.asyncio
async def test_sync_lambda_returning_coroutine_async():
    s = Synchronizer()
    t0 = time.time()
    g = s.create_async(lambda z: f(z + 1))
    coro = g(42)
    assert inspect.iscoroutine(coro)
    assert time.time() - t0 < SLEEP_DELAY
    assert await coro == 1849
    assert time.time() - t0 > SLEEP_DELAY


class Base:
    def __init__(self, x):
        self._x = x


class MyClass(Base):
    def __init__(self, x):
        super().__init__(x)

    async def start(self):
        async def task():
            await asyncio.sleep(SLEEP_DELAY)
            return self._x

        loop = asyncio.get_event_loop()
        self._task = loop.create_task(task())

    async def get_result(self):
        ret = await self._task
        return ret**2

    async def __aenter__(self):
        await asyncio.sleep(SLEEP_DELAY)
        return 42

    async def __aexit__(self, exc_type, exc, tb):
        await asyncio.sleep(SLEEP_DELAY)

    @staticmethod
    async def my_static_method():
        await asyncio.sleep(SLEEP_DELAY)
        return 43

    @classmethod
    async def my_class_method(cls):
        await asyncio.sleep(SLEEP_DELAY)
        return 44

    async def __aiter__(self):
        for i in range(self._x):
            yield i


def test_class_sync():
    s = Synchronizer()
    BlockingMyClass = s.create_blocking(MyClass)
    BlockingBase = s.create_blocking(Base)
    assert BlockingMyClass.__name__ == "BlockingMyClass"
    obj = BlockingMyClass(x=42)
    assert isinstance(obj, BlockingMyClass)
    assert isinstance(obj, BlockingBase)
    obj.start()
    ret = obj.get_result()
    assert ret == 1764

    t0 = time.time()
    with obj as z:
        assert z == 42
        assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY
    assert time.time() - t0 > 2 * SLEEP_DELAY

    t0 = time.time()
    assert BlockingMyClass.my_static_method() == 43
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    t0 = time.time()
    assert BlockingMyClass.my_class_method() == 44
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    assert list(z for z in obj) == list(range(42))


def test_class_sync_futures():
    s = Synchronizer()
    BlockingMyClass = s.create_blocking(MyClass)
    BlockingBase = s.create_blocking(Base)
    assert BlockingMyClass.__name__ == "BlockingMyClass"
    obj = BlockingMyClass(x=42)
    assert isinstance(obj, BlockingMyClass)
    assert isinstance(obj, BlockingBase)
    obj.start()
    fut = obj.get_result(_future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert fut.result() == 1764

    t0 = time.time()
    with obj as z:
        assert z == 42
        assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    assert time.time() - t0 > 2 * SLEEP_DELAY


@pytest.mark.asyncio
async def test_class_async():
    s = Synchronizer()
    AsyncMyClass = s.create_async(MyClass)
    AsyncBase = s.create_async(Base)
    assert AsyncMyClass.__name__ == "AsyncMyClass"
    obj = AsyncMyClass(x=42)
    assert isinstance(obj, AsyncMyClass)
    assert isinstance(obj, AsyncBase)
    await obj.start()
    coro = obj.get_result()
    assert inspect.iscoroutine(coro)
    assert await coro == 1764

    t0 = time.time()
    async with obj as z:
        assert z == 42
        assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY

    assert time.time() - t0 > 2 * SLEEP_DELAY

    lst = []
    async for z in obj:
        lst.append(z)
    assert lst == list(range(42))


@pytest.mark.asyncio
async def test_class_async_back_and_forth():
    s = Synchronizer()
    AsyncMyClass = s.create_async(MyClass)
    AsyncBase = s.create_async(Base)
    s.create_blocking(MyClass)
    async_obj = AsyncMyClass(x=42)
    assert isinstance(async_obj, AsyncMyClass)
    assert isinstance(async_obj, AsyncBase)
    await async_obj.start()

    def get(o):
        return o.get_result()  # Blocking

    # Make it into a sync object
    blocking_obj = s._translate_out(s._translate_in(async_obj), Interface.BLOCKING)
    assert type(blocking_obj).__name__ == "BlockingMyClass"

    # Run it in a sync context
    loop = asyncio.get_event_loop()
    fut = loop.run_in_executor(None, get, blocking_obj)
    ret = await fut
    assert ret == 1764

    # The problem here is that f is already synchronized by another synchronizer, which shouldn't be allowed


@pytest.mark.skip(
    reason="Skip this until we've made it impossible to re-synchronize objects"
)
def test_event_loop():
    s = Synchronizer()
    t0 = time.time()
    f_s = s.create_blocking(f)
    assert f_s(42) == 42 * 42
    assert SLEEP_DELAY < time.time() - t0 < 2 * SLEEP_DELAY
    assert s._thread.is_alive()
    assert s._loop.is_running()
    s._close_loop()
    assert not s._loop.is_running()
    assert not s._thread.is_alive()

    new_loop = asyncio.new_event_loop()
    s._start_loop(new_loop)
    assert s._loop == new_loop
    assert s._loop.is_running()
    assert s._thread.is_alive()

    # Starting a loop again before closing throws.
    with pytest.raises(Exception):
        s._start_loop(new_loop)


@pytest.mark.parametrize("interface_type", [Interface.BLOCKING, Interface.ASYNC])
def test_doc_transfer(interface_type):
    class Foo:
        """Hello"""

        def foo(self):
            """hello"""

    s = Synchronizer()
    output_class = s._wrap(Foo, interface_type)

    assert output_class.__doc__ == "Hello"
    assert output_class.foo.__doc__ == "hello"


def test_set_function_name():
    s = Synchronizer()
    f_s = s.create_blocking(f, "xyz")
    assert f_s(42) == 1764
    assert f_s.__name__ == "xyz"


def test_set_class_name():
    s = Synchronizer()
    BlockingBase = s.create_blocking(Base, "XYZBase")
    assert BlockingBase.__name__ == "XYZBase"
    BlockingMyClass = s.create_blocking(MyClass, "XYZMyClass")
    assert BlockingMyClass.__name__ == "XYZMyClass"



import pickle
import pytest

from synchronicity import Interface, Synchronizer


class PicklableClass:
    async def f(self, x):
        return x**2


@pytest.mark.skip(reason="Let's revisit this in 0.2.0")
def test_pickle():
    s = Synchronizer()
    BlockingPicklableClass = s.create(PicklableClass, Interface.BLOCKING)
    obj = BlockingPicklableClass()
    assert obj.f(42) == 1764
    data = pickle.dumps(obj)
    obj2 = pickle.loads(data)
    assert obj2.f(43) == 1849


def test_pickle_synchronizer():
    s = Synchronizer()
    pickle.dumps(s)



import inspect

from synchronicity import Synchronizer


class _Api:
    def blocking_func(self):
        pass

    async def async_func(self):
        pass


def test_inspect_coroutinefunction():
    s = Synchronizer()
    BlockingApi = s.create_blocking(_Api)
    AioApi = s.create_async(_Api)

    assert inspect.iscoroutinefunction(BlockingApi.blocking_func) is False
    assert inspect.iscoroutinefunction(BlockingApi.async_func) is False
    assert inspect.iscoroutinefunction(AioApi.blocking_func) is False
    assert inspect.iscoroutinefunction(AioApi.async_func) is True



from gevent import monkey

monkey.patch_all()

import asyncio

from synchronicity import Synchronizer


async def f(x):
    await asyncio.sleep(0.1)
    return x**2


s = Synchronizer()
f_s = s.create_blocking(f)
for i in range(3):
    assert f_s(42) == 1764



import asyncio
import pytest
import time

from synchronicity import Interface, Synchronizer


def sleep(ms):
    time.sleep(ms / 1000)
    return ms


async def sleep_async(ms):
    time.sleep(ms / 1000)
    return ms


@pytest.mark.asyncio
async def test_blocking():
    s = Synchronizer()
    sleep_cb = s.create_callback(sleep, Interface.BLOCKING)
    t0 = time.time()
    coros = [sleep_cb(200), sleep_cb(300)]
    rets = await asyncio.gather(*coros)
    assert rets == [200, 300]
    assert 0.3 <= time.time() - t0 <= 0.4  # make sure they run in parallel


@pytest.mark.asyncio
async def test_async():
    s = Synchronizer()
    sleep_cb = s.create_callback(sleep_async, Interface.ASYNC)
    t0 = time.time()
    coros = [sleep_cb(200), sleep_cb(300)]
    rets = await asyncio.gather(*coros)
    assert rets == [200, 300]
    assert 0.3 <= time.time() - t0 <= 0.4


@pytest.mark.asyncio
async def test_translate():
    s = Synchronizer()

    class Foo:
        def __init__(self, x):
            self.x = x

        def get(self):
            return self.x

    BlockingFoo = s.create_blocking(Foo)

    def f(foo):
        assert isinstance(foo, BlockingFoo)
        x = foo.get()
        return BlockingFoo(x + 1)

    f_cb = s.create_callback(f, Interface.BLOCKING)

    foo1 = Foo(42)
    foo2 = await f_cb(foo1)
    assert foo2.x == 43



import asyncio
import atexit
import contextlib
import functools
import inspect
import platform
import threading
import typing
import warnings
from typing import Optional

from .async_wrap import wraps_by_interface
from .callback import Callback
from .exceptions import UserCodeException, unwrap_coro_exception, wrap_coro_exception
from .interface import Interface

_BUILTIN_ASYNC_METHODS = {
    "__aiter__": "__iter__",
    "__aenter__": "__enter__",
    "__aexit__": "__exit__",
}

_RETURN_FUTURE_KWARG = "_future"

# Default names for classes
_CLASS_PREFIXES = {
    Interface.BLOCKING: "Blocking",
    Interface.ASYNC: "Async",
}

# Default names for functions
_FUNCTION_PREFIXES = {
    Interface.BLOCKING: "blocking_",
    Interface.ASYNC: "async_",
}

TARGET_INTERFACE_ATTR = "_sync_target_interface"
SYNCHRONIZER_ATTR = "_sync_synchronizer"


def warn_old_modal_client():
    warnings.warn(
        "Using latest synchronicity with an old interface - please upgrade to latest modal-client!"
    )


class Synchronizer:
    """Helps you offer a blocking (synchronous) interface to asynchronous code."""

    def __init__(
        self,
        multiwrap_warning=False,
        async_leakage_warning=True,
    ):
        self._multiwrap_warning = multiwrap_warning
        self._async_leakage_warning = async_leakage_warning
        self._loop = None
        self._loop_creation_lock = threading.Lock()
        self._thread = None
        self._stopping = None

        if platform.system() == "Windows":
            # default event loop policy on windows spits out errors when
            # closing the event loop, so use WindowsSelectorEventLoopPolicy instead
            # https://stackoverflow.com/questions/45600579/asyncio-event-loop-is-closed-when-getting-loop
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        # Special attribute we use to go from wrapped <-> original
        self._wrapped_attr = "_sync_wrapped_%d" % id(self)
        self._original_attr = "_sync_original_%d" % id(self)

        # Special attribute to mark something as non-wrappable
        self._nowrap_attr = "_sync_nonwrap_%d" % id(self)

        # Prep a synchronized context manager in case one is returned and needs translation
        self._ctx_mgr_cls = contextlib._AsyncGeneratorContextManager
        self.create_async(self._ctx_mgr_cls)
        self.create_blocking(self._ctx_mgr_cls)

        atexit.register(self._close_loop)

    _PICKLE_ATTRS = [
        "_multiwrap_warning",
        "_async_leakage_warning",
    ]

    def __getstate__(self):
        return dict([(attr, getattr(self, attr)) for attr in self._PICKLE_ATTRS])

    def __setstate__(self, d):
        for attr in self._PICKLE_ATTRS:
            setattr(self, attr, d[attr])

    def _start_loop(self):
        with self._loop_creation_lock:
            if self._loop and self._loop.is_running():
                return self._loop

            is_ready = threading.Event()

            def thread_inner():
                async def loop_inner():
                    self._loop = asyncio.get_running_loop()
                    self._stopping = asyncio.Event()
                    is_ready.set()
                    await self._stopping.wait()  # wait until told to stop

                asyncio.run(loop_inner())

            self._thread = threading.Thread(target=thread_inner, daemon=True)
            self._thread.start()
            is_ready.wait()  # TODO: this might block for a very short time
            return self._loop

    def _close_loop(self):
        if self._thread is not None:
            if not self._loop.is_closed():
                # This also serves the purpose of waking up an idle loop
                self._loop.call_soon_threadsafe(self._stopping.set)
            self._thread.join()
            self._thread = None

    def _get_loop(self, start=False):
        if self._loop is None and start:
            return self._start_loop()
        return self._loop

    def _get_running_loop(self):
        # TODO: delete this method
        try:
            return asyncio.get_running_loop()
        except RuntimeError:
            return

    def _is_inside_loop(self):
        loop = self._get_loop()
        if loop is None:
            return False
        if threading.current_thread() != self._thread:
            # gevent does something bad that causes asyncio.get_running_loop() to return self._loop
            return False
        current_loop = self._get_running_loop()
        return loop == current_loop

    def _wrap_check_async_leakage(self, coro):
        """Check if a coroutine returns another coroutine (or an async generator) and warn.

        The reason this is important to catch is that otherwise even synchronized code might end up
        "leaking" async code into the caller.
        """
        if not self._async_leakage_warning:
            return coro

        async def coro_wrapped():
            value = await coro
            # TODO: we should include the name of the original function here
            if inspect.iscoroutine(value):
                warnings.warn(
                    f"Potential async leakage: coroutine returned a coroutine {value}."
                )
            elif inspect.isasyncgen(value):
                warnings.warn(
                    f"Potential async leakage: Coroutine returned an async generator {value}."
                )
            return value

        return coro_wrapped()

    def _wrap_instance(self, obj, interface):
        # Takes an object and creates a new proxy object for it
        cls = obj.__class__
        cls_dct = cls.__dict__
        interfaces = cls_dct[self._wrapped_attr]
        if interface not in interfaces:
            raise RuntimeError(f"Class {cls} has not synchronized {interface}.")
        interface_cls = interfaces[interface]
        new_obj = interface_cls.__new__(interface_cls)
        # Store a reference to the original object
        new_obj.__dict__[self._original_attr] = obj
        new_obj.__dict__[SYNCHRONIZER_ATTR] = self
        new_obj.__dict__[TARGET_INTERFACE_ATTR] = interface
        return new_obj

    def _translate_scalar_in(self, obj):
        # If it's an external object, translate it to the internal type
        if hasattr(obj, "__dict__"):
            if inspect.isclass(obj):  # TODO: functions?
                return obj.__dict__.get(self._original_attr, obj)
            else:
                return obj.__dict__.get(self._original_attr, obj)
        else:
            return obj

    def _translate_scalar_out(self, obj, interface):
        # If it's an internal object, translate it to the external interface
        if inspect.isclass(obj) or isinstance(obj, typing.TypeVar):  # TODO: functions?
            cls_dct = obj.__dict__
            if self._wrapped_attr in cls_dct:
                return cls_dct[self._wrapped_attr][interface]
            else:
                return obj
        else:
            cls_dct = obj.__class__.__dict__
            if self._wrapped_attr in cls_dct:
                # This is an *instance* of a synchronized class, translate its type
                return self._wrap(obj, interface)
            else:
                return obj

    def _recurse_map(self, mapper, obj):
        if type(obj) == list:
            return list(self._recurse_map(mapper, item) for item in obj)
        elif type(obj) == tuple:
            return tuple(self._recurse_map(mapper, item) for item in obj)
        elif type(obj) == dict:
            return dict(
                (key, self._recurse_map(mapper, item)) for key, item in obj.items()
            )
        else:
            return mapper(obj)

    def _translate_in(self, obj):
        return self._recurse_map(self._translate_scalar_in, obj)

    def _translate_out(self, obj, interface):
        return self._recurse_map(
            lambda scalar: self._translate_scalar_out(scalar, interface), obj
        )

    def _translate_coro_out(self, coro, interface):
        async def unwrap_coro():
            return self._translate_out(await coro, interface)

        return unwrap_coro()

    def _run_function_sync(self, coro, interface):
        if self._is_inside_loop():
            raise Exception(
                "Deadlock detected: calling a sync function from the synchronizer loop"
            )

        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)
        fut = asyncio.run_coroutine_threadsafe(coro, loop)
        value = fut.result()
        return self._translate_out(value, interface)

    def _run_function_sync_future(self, coro, interface):
        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)
        # For futures, we unwrap the result at this point, not in f_wrapped
        coro = unwrap_coro_exception(coro)
        coro = self._translate_coro_out(coro, interface)
        return asyncio.run_coroutine_threadsafe(coro, loop)

    async def _run_function_async(self, coro, interface):
        coro = wrap_coro_exception(coro)
        coro = self._wrap_check_async_leakage(coro)
        loop = self._get_loop(start=True)
        if self._is_inside_loop():
            value = await coro
        else:
            c_fut = asyncio.run_coroutine_threadsafe(coro, loop)
            a_fut = asyncio.wrap_future(c_fut)
            value = await a_fut
        return self._translate_out(value, interface)

    def _run_generator_sync(self, gen, interface):
        value, is_exc = None, False
        while True:
            try:
                if is_exc:
                    value = self._run_function_sync(gen.athrow(value), interface)
                else:
                    value = self._run_function_sync(gen.asend(value), interface)
            except UserCodeException as uc_exc:
                raise uc_exc.exc from None
            except StopAsyncIteration:
                break
            try:
                value = yield value
                is_exc = False
            except BaseException as exc:
                value = exc
                is_exc = True

    async def _run_generator_async(self, gen, interface):
        value, is_exc = None, False
        while True:
            try:
                if is_exc:
                    value = await self._run_function_async(gen.athrow(value), interface)
                else:
                    value = await self._run_function_async(gen.asend(value), interface)
            except UserCodeException as uc_exc:
                raise uc_exc.exc from None
            except StopAsyncIteration:
                break
            try:
                value = yield value
                is_exc = False
            except BaseException as exc:
                value = exc
                is_exc = True

    def create_callback(self, f, interface):
        return Callback(self, f, interface)

    def _update_wrapper(
        self, f_wrapped, f, name=None, interface=None, target_module=None
    ):
        """Very similar to functools.update_wrapper"""
        functools.update_wrapper(f_wrapped, f)
        if name is not None:
            f_wrapped.__name__ = name
            f_wrapped.__qualname__ = name
        if target_module is not None:
            f_wrapped.__module__ = target_module
        setattr(f_wrapped, SYNCHRONIZER_ATTR, self)
        setattr(f_wrapped, TARGET_INTERFACE_ATTR, interface)

    def _wrap_callable(
        self,
        f,
        interface,
        name=None,
        allow_futures=True,
        unwrap_user_excs=True,
        target_module=None,
    ):
        if hasattr(f, self._original_attr):
            if self._multiwrap_warning:
                warnings.warn(
                    f"Function {f} is already wrapped, but getting wrapped again"
                )
            return f

        if name is None:
            if hasattr(self, "get_name"):
                # super dumb backwards compatibility fix
                warn_old_modal_client()
                name = self.get_name(f, interface)
            else:
                name = _FUNCTION_PREFIXES[interface] + f.__name__

        is_coroutinefunction = inspect.iscoroutinefunction(f)

        @wraps_by_interface(interface, f)
        def f_wrapped(*args, **kwargs):
            return_future = kwargs.pop(_RETURN_FUTURE_KWARG, False)

            # If this gets called with an argument that represents an external type,
            # translate it into an internal type
            args = self._translate_in(args)
            kwargs = self._translate_in(kwargs)

            # Call the function
            res = f(*args, **kwargs)

            # Figure out if this is a coroutine or something
            is_coroutine = inspect.iscoroutine(res)
            is_asyncgen = inspect.isasyncgen(res)

            if return_future:
                if not allow_futures:
                    raise Exception("Can not return future for this function")
                elif is_coroutine:
                    return self._run_function_sync_future(res, interface)
                elif is_asyncgen:
                    raise Exception("Can not return futures for generators")
                else:
                    return res
            elif is_coroutine:
                if interface == Interface.ASYNC:
                    coro = self._run_function_async(res, interface)
                    if not is_coroutinefunction:
                        # If this is a non-async function that returns a coroutine,
                        # then this is the exit point, and we need to unwrap any
                        # wrapped exception here. Otherwise, the exit point is
                        # in async_wrap.py
                        coro = unwrap_coro_exception(coro)
                    return coro
                elif interface == Interface.BLOCKING:
                    # This is the exit point, so we need to unwrap the exception here
                    try:
                        return self._run_function_sync(res, interface)
                    except UserCodeException as uc_exc:
                        # Used to skip a frame when called from `proxy_method`.
                        if unwrap_user_excs:
                            raise uc_exc.exc from None
                        else:
                            raise uc_exc
            elif is_asyncgen:
                # Note that the _run_generator_* functions handle their own
                # unwrapping of exceptions (this happens during yielding)
                if interface == Interface.ASYNC:
                    return self._run_generator_async(res, interface)
                elif interface == Interface.BLOCKING:
                    return self._run_generator_sync(res, interface)
            else:
                if inspect.isfunction(res) or isinstance(
                    res, functools.partial
                ):  # TODO: HACKY HACK
                    # TODO: this is needed for decorator wrappers that returns functions
                    # Maybe a bit of a hacky special case that deserves its own decorator
                    @wraps_by_interface(interface, res)
                    def f_wrapped(*args, **kwargs):
                        args = self._translate_in(args)
                        kwargs = self._translate_in(kwargs)
                        f_res = res(*args, **kwargs)
                        return self._translate_out(f_res, interface)

                    return f_wrapped

                return self._translate_out(res, interface)

        self._update_wrapper(f_wrapped, f, name, interface, target_module=target_module)
        setattr(f_wrapped, self._original_attr, f)
        return f_wrapped

    def _wrap_proxy_method(synchronizer_self, method, interface, allow_futures=True):
        if getattr(method, synchronizer_self._nowrap_attr, None):
            # This method is marked as non-wrappable
            return method

        method = synchronizer_self._wrap_callable(
            method, interface, allow_futures=allow_futures, unwrap_user_excs=False
        )

        @wraps_by_interface(interface, method)
        def proxy_method(self, *args, **kwargs):
            instance = self.__dict__[synchronizer_self._original_attr]
            try:
                return method(instance, *args, **kwargs)
            except UserCodeException as uc_exc:
                raise uc_exc.exc from None

        return proxy_method

    def _wrap_proxy_staticmethod(self, method, interface):
        orig_function = method.__func__
        method = self._wrap_callable(orig_function, interface)
        self._update_wrapper(method, orig_function, interface=interface)
        return staticmethod(method)

    def _wrap_proxy_classmethod(self, method, interface):
        method = self._wrap_callable(method.__func__, interface)

        @wraps_by_interface(interface, method)
        def proxy_classmethod(cls, *args, **kwargs):
            return method(cls, *args, **kwargs)

        self._update_wrapper(proxy_classmethod, method, interface=interface)
        return classmethod(proxy_classmethod)

    def _wrap_proxy_property(self, prop, interface):
        kwargs = {}
        for attr in ["fget", "fset", "fdel"]:
            if getattr(prop, attr):
                func = getattr(prop, attr)
                kwargs[attr] = self._wrap_proxy_method(func, interface, False)
        return property(**kwargs)

    def _wrap_proxy_constructor(synchronizer_self, cls, interface):
        """Returns a custom __init__ for the subclass."""

        def my_init(self, *args, **kwargs):
            # Create base instance
            args = synchronizer_self._translate_in(args)
            kwargs = synchronizer_self._translate_in(kwargs)
            instance = cls(*args, **kwargs)

            # Register self as the wrapped one
            interface_instances = {interface: self}
            instance.__dict__[synchronizer_self._wrapped_attr] = interface_instances

            # Store a reference to the original object
            self.__dict__[synchronizer_self._original_attr] = instance

        synchronizer_self._update_wrapper(my_init, cls.__init__, interface=interface)
        setattr(my_init, synchronizer_self._original_attr, cls.__init__)
        return my_init

    def _wrap_class(self, cls, interface, name, target_module=None):
        bases = tuple(
            self._wrap(base, interface, require_already_wrapped=(name is not None))
            if base != object
            else object
            for base in cls.__bases__
        )
        new_dict = {self._original_attr: cls}
        if cls is not None:
            new_dict["__init__"] = self._wrap_proxy_constructor(cls, interface)

        for k, v in cls.__dict__.items():
            if k in _BUILTIN_ASYNC_METHODS:
                k_sync = _BUILTIN_ASYNC_METHODS[k]
                if interface == Interface.BLOCKING:
                    new_dict[k_sync] = self._wrap_proxy_method(
                        v, interface, allow_futures=False
                    )
                if interface == Interface.ASYNC:
                    new_dict[k] = self._wrap_proxy_method(
                        v, interface, allow_futures=False
                    )
            elif k in ("__new__", "__init__"):
                # Skip custom constructor in the wrapped class
                # Instead, delegate to the base class constructor and wrap it
                pass
            elif isinstance(v, staticmethod):
                # TODO(erikbern): this feels pretty hacky
                new_dict[k] = self._wrap_proxy_staticmethod(v, interface)
            elif isinstance(v, classmethod):
                new_dict[k] = self._wrap_proxy_classmethod(v, interface)
            elif isinstance(v, property):
                new_dict[k] = self._wrap_proxy_property(v, interface)
            elif callable(v):
                new_dict[k] = self._wrap_proxy_method(v, interface)

        if name is None:
            if hasattr(self, "get_name"):
                # super dumb backwards compatibility fix
                warn_old_modal_client()
                name = self.get_name(cls, interface)
            else:
                name = _CLASS_PREFIXES[interface] + cls.__name__

        new_cls = type.__new__(type, name, bases, new_dict)
        new_cls.__module__ = cls.__module__ if target_module is None else target_module
        new_cls.__doc__ = cls.__doc__
        if "__annotations__" in cls.__dict__:
            new_cls.__annotations__ = cls.__annotations__  # transfer annotations

        setattr(new_cls, TARGET_INTERFACE_ATTR, interface)
        setattr(new_cls, SYNCHRONIZER_ATTR, self)
        return new_cls

    def _wrap(
        self,
        obj,
        interface,
        name=None,
        require_already_wrapped=False,
        target_module=None,
    ):
        # This method works for classes, functions, and instances
        # It wraps the object, and caches the wrapped object

        # Get the list of existing interfaces
        if self._wrapped_attr not in obj.__dict__:
            if isinstance(obj.__dict__, dict):
                # This works for instances
                obj.__dict__.setdefault(self._wrapped_attr, {})
            else:
                # This works for classes & functions
                setattr(obj, self._wrapped_attr, {})

        # If this is already wrapped, return the existing interface
        interfaces = obj.__dict__[self._wrapped_attr]
        if interface in interfaces:
            if self._multiwrap_warning:
                warnings.warn(
                    f"Object {obj} is already wrapped, but getting wrapped again"
                )
            return interfaces[interface]

        if require_already_wrapped:
            # This happens if a class has a custom name but its base class doesn't
            raise RuntimeError(
                f"{obj} needs to be serialized explicitly with a custom name"
            )

        # Wrap object (different cases based on the type)
        if inspect.isclass(obj):
            new_obj = self._wrap_class(
                obj, interface, name, target_module=target_module
            )
        elif inspect.isfunction(obj):
            new_obj = self._wrap_callable(
                obj, interface, name, target_module=target_module
            )
        elif isinstance(obj, typing.TypeVar):
            new_obj = self._wrap_type_var(obj, interface, name, target_module)
        elif self._wrapped_attr in obj.__class__.__dict__:
            new_obj = self._wrap_instance(obj, interface)
        else:
            raise Exception("Argument %s is not a class or a callable" % obj)

        # Store the interface on the obj and return
        interfaces[interface] = new_obj
        return new_obj

    def _wrap_type_var(self, obj, interface, name, target_module):
        # TypeVar translation is needed only for type stub generation, in case the
        # "bound" attribute refers to a translatable type.

        # Creates a new identical TypeVar, marked with synchronicity's special attributes
        # This lets type stubs "translate" the `bounds` attribute on emitted type vars
        # if picked up from module scope and in generics using the base implementation type

        # TODO(elias): Refactor - since this isn't used for live apps, move type stub generation into genstub
        new_obj = typing.TypeVar(name, bound=obj.__bound__)  # noqa
        new_obj.__dict__[self._original_attr] = obj
        new_obj.__dict__[SYNCHRONIZER_ATTR] = self
        new_obj.__dict__[TARGET_INTERFACE_ATTR] = interface
        new_obj.__module__ = target_module
        obj.__dict__.setdefault(self._wrapped_attr, {})
        obj.__dict__[self._wrapped_attr][interface] = new_obj
        return new_obj

    def nowrap(self, obj):
        setattr(obj, self._nowrap_attr, True)
        return obj

    # New interface that (almost) doesn't mutate objects

    def create_blocking(
        self, obj, name: Optional[str] = None, target_module: Optional[str] = None
    ):
        wrapped = self._wrap(obj, Interface.BLOCKING, name, target_module=target_module)
        return wrapped

    def create_async(
        self, obj, name: Optional[str] = None, target_module: Optional[str] = None
    ):
        wrapped = self._wrap(obj, Interface.ASYNC, name, target_module=target_module)
        return wrapped

    def is_synchronized(self, obj):
        if inspect.isclass(obj) or inspect.isfunction(obj):
            return hasattr(obj, self._original_attr)
        else:
            return hasattr(obj.__class__, self._original_attr)

    ### DEPRECATED
    # Only needed because old modal clients don't pin the synchronicity version,
    # so we need this for backwards compatibility for a short while

    def create(self, obj):
        warn_old_modal_client()
        return {
            Interface.ASYNC: self.create_async(obj),
            Interface.BLOCKING: self.create_blocking(obj),
        }

    def __call__(self, obj):
        warn_old_modal_client()
        return self.create_blocking(obj)



import asyncio


class UserCodeException(Exception):
    """This is used to wrap and unwrap exceptions in "user code".

    This lets us have cleaner tracebacks without all the internal synchronicity stuff."""

    def __init__(self, exc):
        # There's always going to be one place inside synchronicity where we
        # catch the exception. We can always safely remove that frame from the
        # traceback.
        self.exc = exc


def wrap_coro_exception(coro):
    async def coro_wrapped():
        try:
            return await coro
        except StopAsyncIteration:
            raise
        except asyncio.CancelledError:
            # we don't want to wrap these since cancelled Task's are otherwise
            # not properly marked as cancelled, and then not treated correctly
            # during event loop shutdown (perhaps in other places too)
            raise
        except UserCodeException:
            raise  # Pass-through in case it got double-wrapped
        except BaseException as exc:
            exc = exc.with_traceback(exc.__traceback__.tb_next)
            raise UserCodeException(exc)

    return coro_wrapped()


async def unwrap_coro_exception(coro):
    try:
        return await coro
    except UserCodeException as uc_exc:
        raise uc_exc.exc from None
